/*
 *  Abstraction routines for MIXER control
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#define ULTRA_MAIN_OBJECT_FILE
#include "driver.h"
#include "mixer.h"

#define ULTRA_MIXERS		(ULTRA_CARDS * 2)

static ultra_mixer_t *ultra_mixers[ ULTRA_MIXERS ];

ultra_mixer_channel_t *ultra_mixer_oss_channel( ultra_mixer_t *mixer, int device )
{
  int idx;
  ultra_mixer_channel_t *channel;
  
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    if ( channel -> hw.ossdev == device )
      return channel;
  }
  return NULL;
}

static int ultra_mixer_open( unsigned short minor, struct file *file )
{
  ultra_mixer_file_t *mfile, *mfile1;
  ultra_mixer_t *mixer;
  unsigned short device = 0;

  if ( minor >= ULTRA_MINOR_UBEGIN ) {
    if ( (minor & ULTRA_MINOR_UDEVMASK) == ULTRA_MINOR_UMIXER1 ) device = 1;
  } else {
    if ( (minor & ULTRA_MINOR_OSS_MASK) == ULTRA_MINOR_MIXER1 ) device = 1;
  }
  minor >>= 3;
  minor &= 0x0e;
  minor |= device;
  if ( minor >= (ultra_cards_count << 1) ) return -ENXIO;
  if ( !(mixer = ultra_mixers[ minor ]) ) return -ENXIO;
  mfile = (ultra_mixer_file_t *)ultra_malloc( sizeof( ultra_mixer_file_t ) );
  if ( !mfile ) return -ENOMEM;
  MEMSET( mfile, 0, sizeof( ultra_mixer_file_t ) );
  SLEEP_PREPARE( mfile, change );
  mfile -> mixer = mixer;
  if ( !mfile -> mixer ) {
    ultra_free( mfile, sizeof( ultra_mixer_file_t ) );
    return -ENOMEM;
  }
  file -> private_data = mfile;
  MOD_INC_USE_COUNT;
  mixer -> card -> use_inc();
  MUTEX_DOWN( mixer, ffile );
  if ( mixer -> ffile ) {
    for ( mfile1 = mixer -> ffile; mfile1 -> next; mfile1 = mfile1 -> next );
    mfile1 -> next = mfile;
  } else {
    mixer -> ffile = mfile;
  }
  MUTEX_UP( mixer, ffile );
  return 0;
}

static int ultra_mixer_release( unsigned short minor, struct file *file )
{
  ultra_mixer_file_t *mfile, *mfile1;
  ultra_mixer_t *mixer;

  if ( file -> private_data ) {
    mfile = (ultra_mixer_file_t *)file -> private_data;
    mixer = mfile -> mixer;
    MUTEX_DOWN( mixer, ffile );
    if ( mixer -> ffile == mfile ) {
      mixer -> ffile = mfile -> next;
    } else {
      for ( mfile1 = mixer -> ffile; mfile1 -> next != mfile; mfile1 = mfile1 -> next );
      mfile1 -> next = mfile -> next;
    }
    MUTEX_UP( mixer, ffile );
    ultra_free( file -> private_data, sizeof( ultra_mixer_file_t ) );
    file -> private_data = NULL;
    mixer -> card -> use_dec();
  }
  MOD_DEC_USE_COUNT;
  return 0;
}

void ultra_mixer_set_kernel_mute( ultra_mixer_t *mixer, unsigned int priority, unsigned short mute ) {
  int idx, mask, left, right;
  ultra_mixer_channel_t *channel;

  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    if ( channel -> hw.priority == priority ) {
      if ( !channel -> hw.stereo ) mute = ULTRA_MIX_MUTE;
      if ( channel -> kmute != mute ) {
        mask = channel -> kmute = mute;
        mask |= channel -> umute;
        if ( mask != channel -> mute ) { 
          if ( channel -> hw.mute && channel -> hw.set_mute ) {
            channel -> hw.set_mute( mixer, channel, mask );
          } else {
            left = mask & ULTRA_MIX_MUTE_LEFT ? 0 : channel -> uleft;
            right = mask & ULTRA_MIX_MUTE_RIGHT ? 0 : channel -> uright;
            if ( channel -> left != left || channel -> right != right ) {
              if ( channel -> hw.set_volume_level )
                channel -> hw.set_volume_level( mixer, channel, left, right );
              channel -> left = left;
              channel -> right = right;
            }        
          }
        }
      }
    }
  }
}

static void ultra_mixer_notify_change( ultra_mixer_file_t *mfile, unsigned int channel )
{
  unsigned long flags;
  ultra_mixer_file_t *mfile1;
  ultra_mixer_t *mixer;
  
  mixer = mfile -> mixer;
  channel = 1 << channel;
  for ( mfile1 = mixer -> ffile; mfile1; mfile1 = mfile1 -> next ) {
    if ( mfile1 != mfile ) {
      CLI( &flags );
      if ( mfile1 -> changes & channel ) {
        STI( &flags );
        continue;
      }
      mfile1 -> changes |= channel;
      STI( &flags );
      if ( GETLOCK( mfile1, change ) & WK_SLEEP ) {
        GETLOCK( mfile1, change ) &= ~WK_SLEEP;
        WAKEUP( mfile1, change );
      }
    }
  }
}

static int ultra_mixer_device_read( ultra_mixer_t *mixer, struct ULTRA_MIXER_CHANNEL *device, int space )
{
  struct ULTRA_MIXER_CHANNEL sdevice, *old_device = NULL;
  ultra_mixer_channel_t *channel;

  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_READ, device, sizeof( *device ) ) ) return -EINVAL;
    MEMCPY_FROMFS( &sdevice, device, sizeof( sdevice ) );
    old_device = device;
    device = &sdevice;
  }
  if ( device -> channel >= mixer -> channels_count ) return -ENODEV;
  channel = mixer -> channels[ device -> channel ];
  device -> flags = 0;
  if ( channel -> record )
    device -> flags |= ULTRA_MIXER_FLG_RECORD;
  if ( channel -> umute & ULTRA_MIX_MUTE_LEFT )
    device -> flags |= ULTRA_MIXER_FLG_MUTE_LEFT;
  if ( channel -> umute & ULTRA_MIX_MUTE_RIGHT )
    device -> flags |= ULTRA_MIXER_FLG_MUTE_RIGHT;
  device -> left = channel -> aleft;
  device -> right = channel -> aright;
  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_WRITE, old_device, sizeof( *old_device ) ) ) return -EINVAL;
    MEMCPY_TOFS( old_device, &sdevice, sizeof( sdevice ) );
  }  
  return 0;
}

static int ultra_mixer_device_write( ultra_mixer_file_t *mfile, struct ULTRA_MIXER_CHANNEL *device, int space )
{
  unsigned int mask, left, right;
  struct ULTRA_MIXER_CHANNEL sdevice, *old_device;
  ultra_mixer_t *mixer;
  ultra_mixer_channel_t *channel;
  int force = 0;

  mixer = mfile -> mixer;
  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_READ, device, sizeof( *device ) ) ) return -EINVAL;
    MEMCPY_FROMFS( &sdevice, device, sizeof( sdevice ) );
    old_device = device;
    device = &sdevice;
  }
  if ( device -> channel >= mixer -> channels_count ) return -ENODEV;
  channel = mixer -> channels[ device -> channel ];
  force = (device -> flags & ULTRA_MIX_FORCE) != 0;
  mask = 0;
  if ( device -> flags & ULTRA_MIXER_FLG_RECORD ) mask = 1;
  if ( force || channel -> record != mask ) {
    if ( channel -> hw.set_record_source )
      channel -> hw.set_record_source( mixer, channel, mask );
    channel -> record = mask;
    ultra_mixer_notify_change( mfile, device -> channel );
  }
  mask = 0;
  if ( channel -> hw.stereo ) {
    if ( device -> flags & ULTRA_MIXER_FLG_MUTE_LEFT ) mask |= ULTRA_MIX_MUTE_LEFT;
    if ( device -> flags & ULTRA_MIXER_FLG_MUTE_RIGHT ) mask |= ULTRA_MIX_MUTE_RIGHT;
  } else {
    if ( device -> flags & ULTRA_MIXER_FLG_MUTE ) mask = ULTRA_MIX_MUTE;
  }
  left = channel -> uleft;
  right = channel -> uright;
  if ( force || channel -> umute != mask ) {
    channel -> umute = mask;
    if ( channel -> hw.mute ) {
      mask |= channel -> kmute;
      channel -> mute = mask;
      if ( channel -> hw.set_mute ) {
        channel -> hw.set_mute( mixer, channel, mask );
      } else {
        if ( mask & ULTRA_MIX_MUTE_LEFT ) left = 0;
        if ( mask & ULTRA_MIX_MUTE_RIGHT ) right = 0;
        if ( force || channel -> left != left || channel -> right != right ) {
          if ( channel -> hw.set_volume_level )
            channel -> hw.set_volume_level( mixer, channel, left, right );
          left = channel -> left;
          right = channel -> right;
        }
      }
      ultra_mixer_notify_change( mfile, device -> channel );
    }
  }
  left = device -> left;
  right = device -> right;
  if ( left > 100 ) left = 100;
  if ( right > 100 ) right = 100;
  if ( force || left != channel -> aleft || right != channel -> aright ) {
    channel -> aleft = left;
    channel -> aright = right;
    if ( channel -> hw.compute_volume ) {
      left = channel -> hw.compute_volume( left );
      right = channel -> hw.compute_volume( right );
    } else {
      left = ( ( channel -> hw.max * left ) + 99 ) / 100;
      right = ( ( channel -> hw.max * right ) + 99 ) / 100;
    }
    channel -> uleft = left;
    channel -> uright = right;
    if ( channel -> hw.mute && channel -> hw.set_mute == NULL ) {
      if ( channel -> mute & ULTRA_MIX_MUTE_LEFT ) left = 0;
      if ( channel -> mute & ULTRA_MIX_MUTE_RIGHT ) right = 0;
    }
    if ( force || channel -> left != left || channel -> right != right ) {
      channel -> left = left;
      channel -> right = right;
      if ( channel -> hw.set_volume_level )
        channel -> hw.set_volume_level( mixer, channel, left, right );
    }
    ultra_mixer_notify_change( mfile, device -> channel );
  }
  return 0;
}

static int ultra_mixer_special_read( ultra_mixer_t *mixer, struct ULTRA_MIXER_SPECIAL *special, int space )
{
  int err = 0;
  struct ULTRA_MIXER_SPECIAL sspecial, *old_special = NULL;
  
  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_READ, special, sizeof( *special ) ) ) return -EINVAL;
    MEMCPY_FROMFS( &sspecial, old_special = special, sizeof( sspecial ) );
    special = &sspecial;
  }
  if ( special -> what == ULTRA_MIXER_S_NONE )
    return 0;		/* can be used for feature testing */
  if ( mixer -> hw.get_special )
    err = mixer -> hw.get_special( mixer, special );
   else
    err = -ENXIO;
  if ( !err && space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_WRITE, old_special, sizeof( *special ) ) ) return -EINVAL;
    MEMCPY_TOFS( old_special, special, sizeof( sspecial ) );
  }
  return err;
}

static int ultra_mixer_special_write( ultra_mixer_t *mixer, struct ULTRA_MIXER_SPECIAL *special, int space )
{
  int err = 0;
  struct ULTRA_MIXER_SPECIAL sspecial;
  
  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_READ, special, sizeof( *special ) ) ) return -EINVAL;
    MEMCPY_FROMFS( &sspecial, special, sizeof( sspecial ) );
    special = &sspecial;
  }
  if ( special -> what == ULTRA_MIXER_S_NONE )
    return 0;		/* can be used for feature testing */
  if ( mixer -> hw.get_special )
    err = mixer -> hw.get_special( mixer, special );
   else
    err = -ENXIO;
  return err;
}

static int ultra_mixer_info( ultra_mixer_t *mixer, struct ULTRA_MIXER_INFO *info, int space )
{
  struct ULTRA_MIXER_INFO sinfo;
  
  sinfo.version = mixer -> card -> version;
  sinfo.caps = mixer -> hw.caps;
  strncpy( sinfo.id, mixer -> id, sizeof( sinfo.id ) );
  strncpy( sinfo.info, mixer -> info, sizeof( sinfo.info ) );
  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_WRITE, info, sizeof( *info ) ) ) return -EINVAL;
    MEMCPY_TOFS( info, &sinfo, sizeof( sinfo ) );
  }  
  return -EIO;
}

static int ultra_mixer_device_info( ultra_mixer_t *mixer, struct ULTRA_MIXER_CHANNEL_INFO *info, int space )
{
  struct ULTRA_MIXER_CHANNEL_INFO sinfo, *old_info = NULL;
  ultra_mixer_channel_t *channel;

  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_READ, info, sizeof( *info ) ) ) return -EINVAL;
    MEMCPY_FROMFS( &sinfo, info, sizeof( sinfo ) );
    old_info = info;
    info = &sinfo;
  }
  if ( info -> channel >= mixer -> channels_count ) return -ENODEV;
  channel = mixer -> channels[ info -> channel ];
  strncpy( info -> name, channel -> hw.name, sizeof( sinfo.name ) );
  info -> name[ sizeof( sinfo.name ) - 1 ] = 0;
  info -> caps = 0;
  if ( channel -> hw.record )
    info -> caps |= ULTRA_MIXER_DINFO_CAP_RECORD;
  if ( channel -> hw.stereo )
    info -> caps |= ULTRA_MIXER_DINFO_CAP_STEREO;
  if ( channel -> hw.mute )
    info -> caps |= ULTRA_MIXER_DINFO_CAP_MUTE;
  if ( space == SP_USER ) {
    if ( VERIFY_AREA( VERIFY_WRITE, old_info, sizeof( *old_info ) ) ) return -EINVAL;
    MEMCPY_TOFS( old_info, &sinfo, sizeof( sinfo ) );
  }  
  return -EIO;
}

static int ultra_mixer_set_recsrc( ultra_mixer_file_t *mfile, int recsrc )
{
  int idx, value;
  ultra_mixer_t *mixer;
  ultra_mixer_channel_t *channel;

  mixer = mfile -> mixer;
  for ( idx = 0; idx < ULTRA_MIXER_OSSDEVS; idx++ ) {
    channel = ultra_mixer_oss_channel( mixer, idx );
    if ( channel == NULL ) continue;
    value = (recsrc & (1 << idx)) ? 1 : 0;
    if ( channel -> record != value ) {
      channel -> record = value;
      if ( channel -> hw.record && channel -> hw.set_record_source )
        channel -> hw.set_record_source( mixer, channel, value );
      ultra_mixer_notify_change( mfile, channel -> channel );
    }
  }
  return 0;
}

static int ultra_mixer_devmask( ultra_mixer_t *mixer )
{
  int result, idx, ossdev;
  
  result = 0;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    ossdev = mixer -> channels[ idx ] -> hw.ossdev;
    if ( ossdev == ULTRA_MIXER_UNKNOWN ) continue;
    result |= (1 << ossdev);
  }
  return result;
}

static int ultra_mixer_stereodevs( ultra_mixer_t *mixer )
{
  int result, idx;
  ultra_mixer_channel_t *channel;
  
  result = 0;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    if ( channel -> hw.ossdev == ULTRA_MIXER_UNKNOWN ) continue;
    if ( channel -> hw.stereo )
      result |= (1 << channel -> hw.ossdev);
  }
  return result;
}

static int ultra_mixer_recmask( ultra_mixer_t *mixer )
{
  int result, idx;
  ultra_mixer_channel_t *channel;
  
  result = 0;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    if ( channel -> hw.ossdev == ULTRA_MIXER_UNKNOWN ) continue;
    if ( channel -> hw.record )
      result |= (1 << channel -> hw.ossdev);
  }
  return result;
}

static int ultra_mixer_get_recsrc( ultra_mixer_t *mixer )
{
  int result, idx;
  ultra_mixer_channel_t *channel;
  
  result = 0;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    if ( channel -> hw.ossdev == ULTRA_MIXER_UNKNOWN ) continue;
    if ( channel -> record )
      result |= (1 << channel -> hw.ossdev);
  }
  return result;
}

static int ultra_mixer_osscaps( ultra_mixer_t *mixer )
{
  int result;

  result = 0;
  if ( mixer -> hw.caps & ULTRA_MIXER_INFO_CAP_EXCL_RECORD )
    result |= ULTRA_MIXER_CAP_EXCL_INPUT;
  return result;
}

static int ultra_mixer_set_volume( ultra_mixer_file_t *mfile, int oss_device, int volume )
{
  int err;
  ultra_mixer_t *mixer;
  ultra_mixer_channel_t *channel;
  struct ULTRA_MIXER_CHANNEL device;

  mixer = mfile -> mixer;
  channel = ultra_mixer_oss_channel( mixer, oss_device );
  if ( !channel ) return -ENXIO;
  device.channel = channel -> channel;
  if ( (err = ultra_mixer_device_read( mixer, &device, SP_KERNEL )) < 0 )
    return err;
  device.left = volume & 0xff;
  device.right = (volume >> 8) & 0xff;
  if ( (err = ultra_mixer_device_write( mfile, &device, SP_KERNEL )) < 0 )
    return err;
  return (unsigned short)device.left |
         ((unsigned short)device.right << 8);
}

static int ultra_mixer_get_volume( ultra_mixer_t *mixer, int oss_device )
{
  int err;
  ultra_mixer_channel_t *channel;
  struct ULTRA_MIXER_CHANNEL device;

  channel = ultra_mixer_oss_channel( mixer, oss_device );
  if ( !channel ) return -ENXIO;
  device.channel = channel -> channel;
  if ( (err = ultra_mixer_device_read( mixer, &device, SP_KERNEL )) < 0 )
    return err;
  return (unsigned short)device.left |
         ((unsigned short)device.right << 8);
}

int ultra_mixer_ioctl_card( ultra_card_t *card, struct file *file, unsigned int cmd, unsigned long arg )
{
  ultra_mixer_file_t *mfile, sfile;
  ultra_mixer_t *mixer;

  if ( card ) {
    int idx;

    /* note: this finds only first mixer for card */
    for ( idx = 0; idx < ULTRA_MIXERS; idx++ ) {
      mixer = ultra_mixers[ idx ];
      if ( mixer && mixer -> card == card ) break;      
    }
    if ( !mixer ) return -ENXIO;
    MEMSET( &sfile, 0, sizeof( sfile ) );
    sfile.mixer = mixer;
    mfile = &sfile;
  } else {
    mfile = (ultra_mixer_file_t *)file -> private_data;
    if ( !mfile ) return -EIO;
    mixer = mfile -> mixer;
  }
  if ( ( ( cmd >> 8 ) & 0xff ) == 'U' ) {
    switch ( cmd ) {
      case ULTRA_MIXER_IOCTL_CHANNELS:
        return IOCTL_OUT( arg, mixer -> channels_count );
      case ULTRA_MIXER_IOCTL_INFO:
        return ultra_mixer_info( mixer, (struct ULTRA_MIXER_INFO *)arg, SP_USER );
      case ULTRA_MIXER_IOCTL_CHANNEL_INFO:
        return ultra_mixer_device_info( mixer, (struct ULTRA_MIXER_CHANNEL_INFO *)arg, SP_USER );
      case ULTRA_MIXER_IOCTL_CHANNEL_READ:
        return ultra_mixer_device_read( mixer, (struct ULTRA_MIXER_CHANNEL *)arg, SP_USER );
      case ULTRA_MIXER_IOCTL_CHANNEL_WRITE:
        return ultra_mixer_device_write( mfile, (struct ULTRA_MIXER_CHANNEL *)arg, SP_USER );
      case ULTRA_MIXER_IOCTL_SPECIAL_READ:
        return ultra_mixer_special_read( mixer, (struct ULTRA_MIXER_SPECIAL *)arg, SP_USER );
      case ULTRA_MIXER_IOCTL_SPECIAL_WRITE:
        return ultra_mixer_special_write( mixer, (struct ULTRA_MIXER_SPECIAL *)arg, SP_USER );
    }
    return -ENXIO;
  }
  if ( ( ( cmd >> 8 ) & 0xff ) != 'M' ) return -ENXIO;
#if 0
  PRINTK( ">> mixer ioctl - cmd = 0x%x\n", cmd );
#endif
  if ( cmd == ULTRA_MIXER_OSS_INFO )
    {
      struct oss_mixer_info info;

      strncpy( info.id, mixer -> id, sizeof( info.id ) );
      strncpy( info.name, mixer -> info, sizeof( info.name ) );
      if ( VERIFY_AREA( VERIFY_WRITE, (void *)arg, sizeof( ultra_mixer_info ) ) ) return -EIO;
      MEMCPY_TOFS( (void *)arg, &ultra_mixer_info, sizeof( ultra_mixer_info ) );
      return 0;
    }
  if ( cmd & IOC_IN )
    switch ( cmd ) {
      case ULTRA_MIXER_SET_RECSRC:
        return IOCTL_OUT( arg, ultra_mixer_set_recsrc( mfile, IOCTL_IN( arg ) ) );
      default:
        return IOCTL_OUT( arg, ultra_mixer_set_volume( mfile, cmd & 0xff, IOCTL_IN( arg ) ) );
    }
   else
  if ( cmd & IOC_OUT )
    switch ( cmd ) {
      case ULTRA_MIXER_DEVMASK:
        return IOCTL_OUT( arg, ultra_mixer_devmask( mixer ) );
      case ULTRA_MIXER_STEREODEVS:
        return IOCTL_OUT( arg, ultra_mixer_stereodevs( mixer ) );
      case ULTRA_MIXER_RECMASK:
        return IOCTL_OUT( arg, ultra_mixer_recmask( mixer ) );
      case ULTRA_MIXER_CAPS:
        return IOCTL_OUT( arg, ultra_mixer_osscaps( mixer ) );
      case ULTRA_MIXER_RECSRC:
        return IOCTL_OUT( arg, ultra_mixer_get_recsrc( mixer ) );
      default:
        return IOCTL_OUT( arg, ultra_mixer_get_volume( mixer, cmd & 0xff ) );
    }
#ifdef ULTRACFG_DEBUG
  PRINTK( "MIXER ERR\n" );
  PRINTK( "  MIXER REQUEST: 0x%x, 0x%x\n", cmd, (int)IOCTL_IN( arg ) );
#endif
  return -ENXIO;
}

static int ultra_mixer_ioctl( struct file *file, unsigned int cmd, unsigned long arg )
{
  return ultra_mixer_ioctl_card( NULL, file, cmd, arg );
}

static long ultra_mixer_read( struct file *file, char *buffer, long count )
{
  unsigned long flags;
  unsigned int device;
  unsigned char buf[ 4 ];
  ultra_mixer_file_t *mfile;
  long size = 0;

  mfile = (ultra_mixer_file_t *)file -> private_data;
  if ( !mfile ) return -EIO;
  if ( count < 4 ) return 0;
  for ( device = 0; device < mfile -> mixer -> channels_count; device++ ) {
    CLI( &flags );
    if ( mfile -> changes & (1 << device) ) {
      mfile -> changes &= ~(1 << device);
      STI( &flags );
      buf[ 0 ] = (unsigned char)device;
      buf[ 1 ] = (unsigned char)(device >> 8);
      buf[ 2 ] = (unsigned char)(device >> 16);
      buf[ 3 ] = (unsigned char)(device >> 24);
      MEMCPY_TOFS( buffer, buf, 4 );
      buffer += 4;
      size += 4;
      count -= 4;
      if ( count < 4 ) return size;
    } else {
      STI( &flags );
    }
  }
  return size;
}

#ifdef ULTRA_POLL
static unsigned int ultra_mixer_poll( struct file *file, poll_table *wait )
{
  unsigned long flags;
  unsigned int mask;
  ultra_mixer_file_t *mfile;
  ultra_mixer_t *mixer;

  mfile = (ultra_mixer_file_t *)file -> private_data;
  if ( !mfile ) return -EIO;
  mixer = mfile -> mixer;
 
  CLI( &flags );
  GETLOCK( mfile, change ) |= WK_SLEEP;
  SLEEP_POLL( mfile, change, wait );
  STI( &flags );
  
  mask = 0;
  if ( mixer -> changes )
    mask |= POLLIN | POLLRDNORM;

  return mask;
}
#else
static int ultra_mixer_select( struct file *file, int sel_type, select_table *wait )
{
  unsigned long flags;
  ultra_mixer_file_t *mfile;
  ultra_mixer_t *mixer;

  mfile = (ultra_mixer_file_t *)file -> private_data;
  if ( !mfile ) return -EIO;  
  mixer = mfile -> mixer;

  switch ( sel_type ) {
    case SEL_IN:
      CLI( &flags );
      if ( !mfile -> changes )
        {
          GETLOCK( mfile, change ) |= WK_SLEEP;
          SLEEPS( mfile, change, wait );
          STI( &flags );
          return 0;
        }
      GETLOCK( mfile, change ) &= ~WK_SLEEP;
      STI( &flags );
      return 1;
    case SEL_OUT:
      break;
    case SEL_EX:
      break;
  }
  return 0;
}
#endif

/*
 *  REGISTRATION PART
 */

ultra_mixer_t *ultra_mixer_new( ultra_card_t *card, char *id, char *info, unsigned short port )
{
  ultra_mixer_t *mixer;
  
  mixer = (ultra_mixer_t *)ultra_malloc( sizeof( ultra_mixer_t ) );
  if ( !mixer ) return NULL;
  MEMSET( mixer, 0, sizeof( ultra_mixer_t ) );
  mixer -> card = card;
  if ( id ) {
    strncpy( mixer -> id, id, sizeof( mixer -> id ) - 1 );
    mixer -> id[ sizeof( mixer -> id ) - 1 ] = 0;
  }
  if ( info ) {
    strncpy( mixer -> info, info, sizeof( mixer -> info ) - 1 );
    mixer -> info[ sizeof( mixer -> info ) - 1 ] = 0;
  }
  mixer -> port = port;
  MUTEX_PREPARE( mixer, ffile );
  return mixer;
}

int ultra_mixer_free( ultra_mixer_t *mixer )
{
  int idx;

  if ( !mixer ) return -EINVAL;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    ultra_free( mixer -> channels[ idx ], sizeof( ultra_mixer_channel_t ) );
  }
  if ( mixer -> private_data && mixer -> private_free )
    mixer -> private_free( mixer -> private_data );
  ultra_free( mixer, sizeof( ultra_mixer_t ) );
  return 0;
}

ultra_mixer_channel_t *ultra_mixer_new_channel( ultra_mixer_t *mixer, struct ULTRA_STRU_MIXER_CHANNEL_HW *hw )
{
  int idx, idx1, priority;
  ultra_mixer_channel_t *channel;

  if ( !mixer || mixer -> channels_count >= ULTRA_MIXER_DEVS ) return NULL;
  channel = (ultra_mixer_channel_t *)ultra_malloc( sizeof( ultra_mixer_channel_t ) );
  if ( !channel ) return NULL;
  MEMSET( channel, 0, sizeof( ultra_mixer_channel_t ) );
  MEMCPY( &channel -> hw, hw, sizeof( struct ULTRA_STRU_MIXER_CHANNEL_HW ) );
  priority = channel -> hw.priority;
  MUTEX_DOWN( mixer, ffile );
  for ( idx = 0; idx < mixer -> channels_count; idx++ )
    if ( mixer -> channels[ idx ] -> hw.priority > priority ) break;
  for ( idx1 = mixer -> channels_count; idx1 > idx; idx1-- ) {
    mixer -> channels[ idx1 ] = mixer -> channels[ idx1 - 1 ];
  }
  mixer -> channels_count++;
  mixer -> channels[ idx ] = channel;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    mixer -> channels[ idx ] -> channel = idx;
  }
  MUTEX_UP( mixer, ffile );
  return channel;
}

void ultra_mixer_reorder_channel( ultra_mixer_t *mixer, ultra_mixer_channel_t *channel )
{
  int idx, idx1;

  if ( !mixer || !channel ) return;
  MUTEX_DOWN( mixer, ffile );
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    if ( mixer -> channels[ idx ] == channel ) {
      for ( idx1 = idx; idx1 + 1 < mixer -> channels_count; idx1++ )
        mixer -> channels[ idx1 ] = mixer -> channels[ idx1 + 1 ];
      mixer -> channels_count--;
      break;
    }
  }
  for ( idx = 0; idx < mixer -> channels_count; idx++ )
    if ( mixer -> channels[ idx ] -> hw.priority > channel -> hw.priority ) break;
  for ( idx1 = mixer -> channels_count; idx1 > idx; idx1-- ) {
    mixer -> channels[ idx1 ] = mixer -> channels[ idx1 - 1 ];
  }
  mixer -> channels[ idx ] = channel;
  mixer -> channels_count++;
  MUTEX_UP( mixer, ffile );
}

int ultra_mixer_register( ultra_mixer_t *mixer, int device )
{
  int idx;

  if ( device < 0 || device > 1 ) return -EINVAL;
  if ( !mixer -> card ) return -EINVAL;
  idx = (mixer -> card -> number << 1) + device;
  if ( ultra_mixers[ idx ] ) return -EBUSY;
  ultra_mixers[ idx ] = mixer;
  return 0;
}

int ultra_mixer_unregister( ultra_mixer_t *mixer )
{
  int idx;

  for ( idx = 0; idx < ULTRA_MIXERS; idx++ ) {
    if ( ultra_mixers[ idx ] == mixer ) {
      ultra_mixers[ idx ] = NULL;
    }
  }
  return ultra_mixer_free( mixer );
}

/*
 *  INIT PART
 */

int ultra_mixer_channel_init( ultra_mixer_t *mixer, unsigned int priority, unsigned char left, unsigned char right, unsigned int flags )
{
  int idx, err;
  ultra_mixer_file_t sfile, *mfile;
  ultra_mixer_channel_t *channel;
  struct ULTRA_MIXER_CHANNEL device;

  MEMSET( &sfile, 0, sizeof( sfile ) );
  sfile.mixer = mixer;
  mfile = &sfile;
  for ( idx = 0; idx < mixer -> channels_count; idx++ ) {
    channel = mixer -> channels[ idx ];
    if ( channel -> hw.priority == priority ) {
      device.channel = channel -> channel;
      device.flags = flags | ULTRA_MIX_FORCE;
      device.left = (unsigned char)left;
      device.right = (unsigned char)right;
      if ( (err = ultra_mixer_device_write( mfile, &device, SP_KERNEL )) < 0 )
        return err;
      return 0;
    }
  }
  return -ENODEV;
}

static int ultra_mixer_unregister_callback( void );

static struct ultra_stru_register_minor ultra_mixer_reg = {
  "mixer",

  1,					/* verify soundcard # */

  { ULTRA_MINOR_UMIXER0, ULTRA_MINOR_UMIXER1, -1 },
  { ULTRA_MINOR_MIXER, ULTRA_MINOR_MIXER1, -1 },

  ultra_mixer_unregister_callback,	/* unregister */

  NULL,					/* lseek */
  ultra_mixer_read,			/* read */
  NULL,					/* write */
  ultra_mixer_open,			/* open */
  ultra_mixer_release,			/* release */
#ifdef ULTRA_POLL
  ultra_mixer_poll,			/* poll */
#else
  ultra_mixer_select,			/* select */
#endif
  ultra_mixer_ioctl,			/* ioctl */
  NULL					/* mmap */
};

static int ultra_mixer_unregister_callback( void )
{
  int err;

  if ( ( err = ultra_unregister_minor( &ultra_mixer_reg ) ) < 0 )
    PRINTD( "ultra-mixer: unregister error %d\n", err );
  return err;
}
 
int init_module( void )
{
  int err;
  
  MEMSET( &ultra_mixers, 0, sizeof( ultra_mixers ) );
  if ( ( err = ultra_register_minor( &ultra_mixer_reg ) ) < 0 )
    return err;
  return 0;
}

void cleanup_module( void )
{
  ultra_mixer_unregister_callback();
}
